(define-module (alterator d)
  :use-module (vhttpd)
  :use-module (srfi srfi-1)
  :use-module (alterator algo)
  :use-module (alterator pipe)
  :use-module (alterator woo)
  :use-module (alterator exit-handler)
  :use-module (alterator systemd)
  :export (d-loop
           d-control
           d-query

           d-init-local
           d-init-global
           d-end

           d-wait
           d))

(define *d-user* (passwd:name (getpwuid (getuid))))

; /var/run/alteratord is also used in installer
; to bounce the socket between stage3 and stage2;
; it's also more secure than predefined /tmp subdir
; NB: a copy contained in ../../sbin/alteratord
(define *tmpdir*
  (if (string=? *d-user* "root")
      "/var/run/alteratord"
      (string-append
       (or (getenv "TMPDIR") "/tmp")
       "/alterator")))

(define *d-socket* (string-append *tmpdir* "/.socket"))
(define *d-group* (if (string=? *d-user* "root")
                      "_alteratord"
                      *d-user*))
(define *d-control* "/etc/rc.d/init.d/alteratord")
(define *d-process* #f)
(define *d-restart-service* #f)

(define *null-port* (open-output-file "/dev/null"))

(or (access? *tmpdir* W_OK) (mkdir *tmpdir*))

(define (d-socket)
  (or (systemctl-simple "alteratord" 'socket)
      *d-socket*))

;; conversion between text and s-exp

(define (text->scm str)
  (catch
    #t
    (lambda()
      (with-input-from-string str read))
    (lambda args '())))

(define (scm->text scm)
  (catch
    #t
    (lambda()
      (with-output-to-string (lambda() (write scm))))
    (lambda args "")))

;; API

(define (d-loop proc)
  (let* ((fds (sd-listen-fds))
         (server (if (and fds (not (null? fds)))
                   (begin
                     (format (current-error-port) "*** Make server from the given socket ***~%")
                     (force-output (current-error-port))
                     (make-server-from-socket (car fds)))
                   (begin
                     (format (current-error-port) "*** Make a new server socket ***~%")
                     (force-output (current-error-port))
                     (make-unix-server *d-socket*))))
         (uid (passwd:uid (getpwnam *d-user*)))
         (gid (group:gid (getgrnam *d-group*))))
    (if (or (not fds) (null? fds))
      (begin
        (chmod *d-socket* #o660)
        (chown *d-socket* uid gid)))
    (server-loop server
      (lambda (code request)
        (make-string-response code "text/plain"
          (if (= code 200)
            (scm->text (proc (text->scm (message-body request))))
            "error"))))))

(define (d-control-sysv . args)
  (let ((pid (primitive-fork)))
    (cond
      ((zero? pid)
       (catch/ignore
        #t
        (lambda()
          (dup2 (fileno *null-port*) 1)
          (dup2 (fileno *null-port*) 2)
          (apply execl *d-control* *d-control* args)))
       (primitive-exit 1))
      (else
       (waitpid pid)))))

(define (d-control . args)
  (let ((sval (systemctl-simple "alteratord" 'status)))
    (if (and sval (pair? sval))
      (cons 0 (if (or (null? args)
                      (case (string->symbol (car args))
                        ((status) (eq? 'active (cdr sval)))
                        ((start) (systemctl-simple "alteratord" 'activate))
                        ((stop) (systemctl-simple "alteratord" 'deactivate))
                        (else #f)))
                0
                256))
      (apply d-control-sysv args))))

(define (d-query cmd)
  (request-unix-server (d-socket)
                       (scm->text cmd)
                       (lambda(code response)
                         (if (= code 200)
                             (unpack-exception (text->scm (message-body response)))
                             (error "Unable to connect to alteratord service")))))

(define (d-service-running?)
  (= (status:exit-val (cdr (d-control "status"))) 0))

;;note: don't use control script, because init script pass all stderr to initlog

(define (d-wait)
  (or (= (request-unix-server (d-socket) "" (lambda(code response) code)) 200)
      (begin (usleep 5000)
             (d-wait))))

(define (d-init-local)
  (set! *d-restart-service* (d-service-running?))
  (d-control "stop")
  (set! *d-process* (create-process #f "/usr/sbin/alteratord" "-l"))
  (d-wait))

(define (d-init-global)
  (if (not (d-service-running?))
    (begin
      (set! *d-restart-service* #f)
      (d-control "start")))
  (d-wait))

(define (d-end)
  (if *d-process*
    (begin
      (stop-process 'terminate *d-process*)
      (set! *d-process* #f)))
  (if *d-restart-service*
    (begin
      (d-control "start")
      (set! *d-restart-service* #f))))

;;backward compatibility: woobus-module
(define (d)
  (lambda(cmds next)
    (next (append-map d-query cmds))))

;;; at-exit cleanup
(at-exit d-end)
